import pytest
from pathlib import Path
from dagshub.streaming import DagsHubFilesystem


@pytest.mark.parametrize(
    "path,expected",
    [
        ("regular/path/in/repo", False),
        (".", False),
        (".git/", True),
        (".git/and/then/some", True),
        (".dvc/", True),
        (".dvc/and/then/some", True),
        ("repo/file.dvc/", False),
        ("repo/file.git/", False),
        ("venv/lib/site-packages/some-package", True),
    ],
)
def test_passthrough_path(path, expected):
    path = Path(path)
    actual = DagsHubFilesystem._passthrough_path(path)
    assert actual == expected



import os.path

from httpx import Response
from respx import MockRouter, Route

BASE_REGEX = r"/api/v1/repos/\w+/\w+"


class MockApi(MockRouter):
    def __init__(self, git_repo, user="user", reponame="repo", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user = user
        self.reponame = reponame
        self.git_repo = git_repo

        self._endpoints, self._responses = self._default_endpoints_and_responses()
        route_dict = {
            k: (self._endpoints[k], self._responses[k]) for k in self._endpoints
        }
        for route_name in route_dict:
            endpoint_regex, return_value = route_dict[route_name]
            self.route(name=route_name, url__regex=endpoint_regex).mock(return_value)

    @property
    def repourlpath(self):
        return f"{self.user}/{self.reponame}"

    @property
    def repophysicalpath(self):
        return str(self.git_repo.workspace)

    @property
    def current_revision(self):
        heads = self.git_repo.api.heads
        if "main" in heads:
            return heads.main.commit.hexsha
        else:
            return heads.master.commit.hexsha

    @property
    def api_list_path(self):
        return f"/api/v1/repos/{self.repourlpath}/content/{self.current_revision}"

    @property
    def api_raw_path(self):
        return f"/api/v1/repos/{self.repourlpath}/raw/{self.current_revision}"

    def _default_endpoints_and_responses(self):
        endpoints = {
            "branch": rf"{BASE_REGEX}/branches/\w+",
            "branches": rf"{BASE_REGEX}/branches",
            "list_root": rf"{BASE_REGEX}/content/{self.current_revision}/$",
        }

        responses = {
            "branch": Response(
                200,
                json={
                    "name": "main",
                    "commit": {
                        "id": self.current_revision,
                        "message": "Update 'README.md'\n",
                        "url": "",
                        "author": {
                            "name": "dagshub",
                            "email": "info@dagshub.com",
                            "username": "",
                        },
                        "committer": {
                            "name": "dagshub",
                            "email": "info@dagshub.com",
                            "username": "",
                        },
                        "added": None,
                        "removed": None,
                        "modified": None,
                        "timestamp": "2021-08-10T09:03:32Z",
                    },
                },
            ),
            "branches": Response(
                200,
                json=[
                    {
                        "name": "main",
                        "commit": {
                            "id": self.current_revision,
                            "message": "Update 'README.md'\n",
                            "url": "",
                            "author": {
                                "name": "dagshub",
                                "email": "info@dagshub.com",
                                "username": "",
                            },
                            "committer": {
                                "name": "dagshub",
                                "email": "info@dagshub.com",
                                "username": "",
                            },
                            "added": None,
                            "removed": None,
                            "modified": None,
                            "timestamp": "2021-08-10T09:03:32Z",
                        },
                    }
                ],
            ),
            "list_root": Response(
                200,
                json=[
                    {
                        "path": "a.txt",
                        "type": "file",
                        "size": 0,
                        "hash": "some_hash",
                        "versioning": "dvc",
                        "download_url": "some_url",
                    },
                    {
                        "path": "b.txt",
                        "type": "file",
                        "size": 0,
                        "hash": "some_hash",
                        "versioning": "dvc",
                        "download_url": "some_url",
                    },
                    {
                        "path": "c.txt",
                        "type": "file",
                        "size": 0,
                        "hash": "some_hash",
                        "versioning": "dvc",
                        "download_url": "some_url",
                    },
                    {
                        "path": "a.txt.dvc",
                        "type": "file",
                        "size": 0,
                        "hash": "some_hash",
                        "versioning": "git",
                        "download_url": "some_url",
                    },
                ],
            ),
        }

        return endpoints, responses

    def add_file(self, path, content="aaa", status=200) -> Route:
        """
        Add a file to the api (only accessible via the raw endpoint)
        """
        route = self.route(url=f"{self.api_raw_path}/{path}")
        route.mock(Response(status, content=content))
        return route

    def add_dir(self, path, contents=[], status=200) -> Route:
        """
        Add a directory to the api (only accessible via the content endpoint)
        We don't keep a tree of added dirs, so it's not dynamic
        """
        route = self.route(url=f"{self.api_list_path}/{path}")
        content = [
            self.generate_list_entry(os.path.join(path, c[0]), c[1]) for c in contents
        ]
        route.mock(Response(status, json=content))
        return route

    def generate_list_entry(self, path, type="file"):
        return {
            "path": path,
            "type": type,
            "size": 0,
            "hash": "8586da76f372efa83d832a9d0e664817.dir",
            "versioning": "dvc",
            "download_url": f"https://dagshub.com/{self.repourlpath}/raw/{self.current_revision}/{path}",
        }



import json
import posixpath

import urllib
import os
import logging
import fnmatch
from typing import Union
from io import IOBase
import httpx
import rich.progress

from dagshub.common import config, helpers, rich_console
from http import HTTPStatus
import dagshub.auth
from dagshub.auth.token_auth import HTTPBearerAuth
from dagshub.upload.errors import determine_upload_api_error
from dagshub.common.helpers import log_message

# todo: handle api urls in common package
CONTENT_UPLOAD_URL = "api/v1/repos/{owner}/{reponame}/content/{branch}/{path}"
DEFAULT_COMMIT_MESSAGE = "Upload files using DagsHub client"
DEFAULT_DATASET_COMMIT_MESSAGE = "Initial dataset commit"
REPO_CREATE_URL = "api/v1/user/repos"
ORG_REPO_CREATE_URL = "api/v1/org/{orgname}/repos"
USER_INFO_URL = "api/v1/user"
DEFAULT_DATA_DIR_NAME = 'data'
logger = logging.getLogger(__name__)

s = httpx.Client()
s.timeout = config.http_timeout
s.follow_redirects = True
s.headers.update(config.requests_headers)


def create_dataset(repo_name, local_path, glob_exclude="", org_name="", private=False):
    """
    Create a new repository on DagsHub and upload an entire dataset to it
    :param repo_name (str): Name of the repository to be created
    :param local_path (str): local path where the dataset to upload is located
    :param glob_exclude (str): regex to exclude certain files from the upload process
    :param org_name (str): Organization name to be the repository owner
    :param private (bool): Flag to indicate the repository is going to be private
    :return : Repo object of the repository created
    """
    repo = create_repo(repo_name, org_name=org_name, private=private)
    dir = repo.directory(repo_name)
    dir.add_dir(local_path, glob_exclude, commit_message=DEFAULT_DATASET_COMMIT_MESSAGE)
    return repo


def add_dataset_to_repo(repo,
                        local_path,
                        data_dir=DEFAULT_DATA_DIR_NAME):
    """
    Given a repository created on dagshub - upload an entire dataset to it
    :param reo (Reop): repository created beforehand
    :param local_path (str): local path where the dataset to upload is located
    :param data_dir (str): name of data directory that will be created inside repo
    """
    dir = repo.directory(data_dir)
    dir.add_dir(local_path, commit_message=DEFAULT_DATASET_COMMIT_MESSAGE)


def create_repo(
    repo_name,
    org_name="",
    description="",
    private=False,
    auto_init=False,
    gitignores="Python",
    license="",
    readme="",
    template="custom",
    host=""
):
    """
    Creates a repository on DagsHub for the current user (default) or an organization passed as an argument

    :param repo_name (str): Name of the repository to be created
    :param org_name (str): Organization name to be the repository owner
    :param description (str): Description for the repository
    :param private (bool): Flag to indicate the repository is going to be private
    :param gitignores (str): Which gitignore template(s) to use (comma separated string)
    :param license (str): Which license file to use
    :param readme (str): Readme file path to upload
    :param template (str): Which project template to use, options are: none, custom, notebook-template,
    cookiecutter-dagshub-dvc. To learn more, check out https://dagshub.com/docs/feature_guide/project_templates/
    :return: Repo object of the repository created
    """
    if template == "":
        template = "none"

    import dagshub.auth
    from dagshub.auth.token_auth import HTTPBearerAuth

    host = host or config.host

    username = config.username
    password = config.password
    if username is not None and password is not None:
        auth = username, password
    else:
        token = config.token or dagshub.auth.get_token()
        if token is not None:
            auth = HTTPBearerAuth(token)

    if auth is None:
        raise RuntimeError("You can't create a repository without being authenticated.")

    if (license != "" or readme != "" or gitignores != "") and template == "none":
        template = "custom"

    data = {
        "name": repo_name,
        "description": description,
        "private": private,
        "auto_init": auto_init,
        "gitignores": gitignores,
        "license": license,
        "readme": readme,
        "project_template": template,
    }

    url = REPO_CREATE_URL
    if org_name and not org_name.isspace():
        url = ORG_REPO_CREATE_URL.format(
            orgname=org_name,
        )

    res = s.post(urllib.parse.urljoin(host, url), data=data, auth=auth)

    if res.status_code != HTTPStatus.CREATED:
        logger.error(f"Response ({res.status_code}):\n" f"{res.content}")
        if res.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
            raise RuntimeError("Repository name is invalid or it already exists.")
        else:
            raise RuntimeError("Failed to create the desired repository.")

    repo = res.json()
    return Repo(
        owner=repo["owner"]["login"], name=repo["name"], token=token, branch="main"
    )


class Repo:
    def __init__(
        self, owner, name, username=None, password=None, token=None, branch=None
    ):

        """
        Repo class constructor. If branch is not provided, then default branch is used.

        :param owner (str): Store the username of the user who owns this repository
        :param name (str): Identify the repository
        :param username (str): Set the username to none if it is not provided
        :param password (str): Set the password to none if it is not provided
        :param token (str): Set the token
        :param branch (str): Set the branch to the default branch
        :return: The object of the class

        """
        self.owner = owner
        self.name = name
        self.src_url = config.host

        self.username = username or config.username
        self.password = password or config.password
        self.token = token or config.token
        self.branch = branch

        if self.branch is None:
            logger.debug("Branch wasn't provided. Fetching default branch...")
            self._set_default_branch()
        logger.debug(f"Set branch: {self.branch}")

    def upload(
        self,
        local_path: Union[str, IOBase],
        commit_message=DEFAULT_COMMIT_MESSAGE,
        remote_path=None,
        **kwargs,
    ):
        """
        The upload function is used to upload files to the repository.
        It takes a file as an argument and logs the response status code and content.


        :param file (str): Specify the file to be uploaded
        :param commit_message (str): Specify a commit message
        :param path (str): Specify the path to upload the file to
        :param **kwargs: Pass in any additional parameters that are required for the upload function
        :return: None

        """
        if os.path.isdir(local_path):
            dir_to_upload = self.directory(remote_path)
            dir_to_upload.add_dir(local_path, commit_message=commit_message, **kwargs)
        else:
            file_to_upload = DataSet.get_file(local_path, remote_path)
            self.upload_files([file_to_upload], commit_message=commit_message, **kwargs)

    def upload_files(
        self,
        files,
        directory_path="",
        commit_message=DEFAULT_COMMIT_MESSAGE,
        versioning=None,
        new_branch=None,
        last_commit=None,
        force=False,
    ):
        """
        The upload_files function uploads a list of files to the specified directory.


        :param files (list(str)): Pass the files that are to be uploaded
        :param directory_path (str): Indicate the path of the directory in which we want to upload our files
        :param commit_message (str): Set the commit message
        :param versioning (str): Determine whether the files are uploaded to a new branch or not
        :param new_branch (str): Create a new branch
        :param last_commit (str): Tell the server that we want to upload a file without committing it
        :param force (bool): Force the upload of a file even if it is already present on the server
        :return: None
        """

        data = {
            "commit_choice": "direct",
            "commit_summary": commit_message,
            "versioning": versioning,
            "last_commit": last_commit,
            "is_dvc_dir": directory_path != "" and versioning != "git",
        }

        if new_branch is not None:
            data.update(
                {
                    "commit_choice": "commit-to-new-branch",
                    "new_branch_name": new_branch,
                }
            )

        if force:
            data["last_commit"] = self._get_last_commit()

        log_message(f'Uploading files ({len(files)}) to "{self.full_name}"...', logger)
        res = s.put(
            self.get_request_url(directory_path),
            data=data,
            files=[("files", file) for file in files],
            auth=self.auth,
            timeout=None,
        )
        self._log_upload_details(data, res, files)

    def _log_upload_details(self, data, res, files):
        """
        The _log_upload_details function logs the request URL, data, and files.
        It then logs the response status code and content. If the response is not 200(OK), it raises an error.



        :param data (str): Pass the data that will be uploaded to the server
        :param res (dict): Store the response from the server
        :param files (list(str)): Pass the files that are going to be uploaded
        :return: None

        """

        logger.debug(
            f"Request URL: {res.request.url}\n"
            f"Data:\n{json.dumps(data, indent=4)}\n"
            f"Files:\n{json.dumps(list(map(str, files)), indent=4)}"
        )

        if res.status_code != HTTPStatus.OK:
            raise determine_upload_api_error(res)
        else:
            log_message("Upload finished successfully!", logger)

    @property
    def auth(self):
        """
        The auth function is used to authenticate the user with the dagshub API.
            It takes in a username and password, or token as arguments. If both are provided,
            it will use the username and password combination to
            get a token from dagshub's authentication server.
            Otherwise, if only a token is provided it will be used for authentication.

        :return: The HTTPAuth object

        """
        if self.username is not None and self.password is not None:
            return httpx.BasicAuth(self.username, self.password)
        token = self.token or dagshub.auth.get_token()
        return HTTPBearerAuth(token)

    def directory(self, path):
        """
        The directory function returns a DataSet object that represents the directory at the given path.


        :param path (str): Specify the directory that will contain the data.
                           This directory is the "root" of the dataset.
        :return: A dataset object that represents the directory at the given path

        """
        return DataSet(self, path)

    def get_request_url(self, directory):
        """
        The get_request_url function returns the URL for uploading a file to DagsHub.

        :param directory (str): the path to a directory within this repo on DagsHub.
            For example, if you have created your repo in such a
            way that it has two directories named data and models,
            then you could pass one of these strings into this function as an argument.
        :return: The url for uploading a file

        """
        return urllib.parse.urljoin(
            self.src_url,
            CONTENT_UPLOAD_URL.format(
                owner=self.owner,
                reponame=self.name,
                branch=self.branch,
                path=urllib.parse.quote(directory, safe=""),
            ),
        )

    def _set_default_branch(self):

        """
        The _set_default_branch function is used to set the default branch for a repository.
        It first tries to get the default branch from DagsHub. If it fails, an error message is raised.

        :return: None
        """

        try:
            self.branch = helpers.get_default_branch(
                self.owner, self.name, self.auth, self.src_url
            )
        except Exception:
            raise RuntimeError(
                "Failed to get default branch for repository. "
                "Please specify a branch and make sure repository details are correct."
            )

    @property
    def full_name(self):
        """
        The full_name function returns the full name of a repository, meaning: "<owner>/<repo>"


        :return: The full name of a repository

        """

        return f"{self.owner}/{self.name}"

    def _get_last_commit(self):
        """
        The _get_last_commit function returns the last commit sha for a given branch.
        It is used to check if there are any new commits in the repo since we last ran our dag.

        :return: The commit id of the last commit for the branch
        """
        api_path = f"api/v1/repos/{self.full_name}/branches/{self.branch}"
        api_url = urllib.parse.urljoin(self.src_url, api_path)
        res = s.get(api_url, auth=self.auth)
        if res.status_code == HTTPStatus.OK:
            content = res.json()
            try:
                return content["commit"]["id"]
            except KeyError:
                logger.error(f"Cannot get commit sha for branch '{self.branch}'")
        return ""


class DataSet:
    def __init__(self, repo: Repo, directory: str):
        """

        :param repo (Repo object): Pass a repo object
        :param directory (str): Specify the directory of the repository
        :return: A DataSet object

        """

        self.files = {}
        self.repo = repo
        self.directory = self._clean_directory_name(directory)
        self.request_url = self.repo.get_request_url(directory)

    def add(self, file: Union[str, IOBase], path=None):
        """
        The add function adds a file to the list of files that will be uploaded.


        :param file (str): Specify the file to be uploaded
        :param path (str): Specify the path to upload the file
        :return: None

        """

        path, file = self.get_file(file, path)
        if file is not None:
            if path in self.files:
                log_message(
                    f'File already staged for upload on path "{path}". Overwriting', logger
                )
            self.files[path] = (path, file)

    def add_dir(self, local_path, glob_exclude="", commit_message=None, **upload_kwargs):
        """
        The add_dir function adds an entire dvc directory to a DagsHub repository.
        It does this by iterating through all the files in the given directory and uploading them one-by-one.
        The function also commits all of these changes at once, so as not to overload the API with requests.


        :param local_path  (str): Specify the local path where the dataset to upload is located
        :param glob_exclude (str): Exclude certain files from the upload process
        :param commit_message (str): Commit message
        :param upload_kwargs (dict): kwargs that are passed to the uploading function
        :return: None

        """

        file_counter = 0

        progress = rich.progress.Progress(rich.progress.SpinnerColumn(), *rich.progress.Progress.get_default_columns(),
                                          console=rich_console, transient=True, disable=config.quiet)
        total_task = progress.add_task("Uploading files...")

        with progress:
            for root, dirs, files in os.walk(local_path):
                folder_task = progress.add_task(f"Uploading files from {root}", total=len(files))

                if commit_message is None:
                    commit_message = upload_kwargs.get("commit_message", f"Commit data points in folder {root}")
                if "commit_message" in upload_kwargs:
                    del upload_kwargs["commit_message"]

                if len(files) > 0:
                    for filename in files:
                        rel_file_path = posixpath.join(root, filename)
                        rel_remote_file_path = rel_file_path.replace(local_path, "")
                        if (
                            glob_exclude == ""
                            or fnmatch.fnmatch(rel_file_path, glob_exclude) is False
                        ):
                            self.add(file=rel_file_path, path=rel_remote_file_path)
                            if len(self.files) > 49:
                                file_counter += len(self.files)
                                self.commit(commit_message, versioning="dvc", **upload_kwargs)
                                progress.update(folder_task, advance=len(self.files))
                                progress.update(total_task, completed=file_counter)
                    if len(self.files) > 0:
                        file_counter += len(self.files)
                        self.commit(commit_message, versioning="dvc", **upload_kwargs)
                        progress.update(total_task, completed=file_counter)
                progress.remove_task(folder_task)

        log_message(f"Directory upload complete, uploaded {file_counter} files", logger)

    @staticmethod
    def _clean_directory_name(directory: str):
        """
        The _clean_directory_name function takes a directory name as an argument
        and returns the normalized path of that directory.
        For example, if the input is ../../../ then it will return /. If the input is ./ then it will return ..
        If there are any other characters in the string,
        they will be ignored and only alphanumeric characters (a-zA-Z0-9)
        will be kept.

        :param directory (str): Specify the directory that will be cleaned
        :return: The normalized path of the directory
                 (The directory name with the path separator normalized to a forward slash)

        """

        return posixpath.normpath(directory)

    @staticmethod
    def get_file(file: Union[str, IOBase], path=None):
        """
        The get_file function is a helper function that takes in either a string or an IOBase object and returns
        a tuple containing the file's name and the file itself. If no path is provided, it will default to the name of
        the file.

        :param file (str):Union[str: Specify the file that you want to upload
        :param IOBase]: Handle both file paths and file objects
        :param path (str): Specify the path of the file
        :return: A tuple of the path and a file object

        """

        try:
            # if path is not provided, fall back to the file name
            if path is None:
                try:
                    path = posixpath.basename(
                        posixpath.normpath(file if type(file) is str else file.name)
                    )
                except Exception:
                    raise RuntimeError(
                        "Could not interpret your file's name. Please specify it in the keyword parameter 'path'."
                    )

            if type(file) is str:
                try:
                    f = open(file, "rb")
                    return path, f
                except IsADirectoryError:
                    raise IsADirectoryError(
                        "'file' must describe a file, not a directory."
                    )

            return path, file

        except Exception as e:
            logger.error(e)
            raise e

    def _reset_dataset(self):
        """
        The _reset_dataset function clears the files attribute of a Dataset object.

        :return: None
        """

        self.files.clear()

    def commit(self, commit_message=DEFAULT_COMMIT_MESSAGE, *args, **kwargs):
        """
        The commit function is used to commit the files in the dataset.
        It takes a commit message as an argument,
            if no argument is passed then it return default commit message "Upload files using DagsHub client".
        The function returns nothing.


        :param commit_message (str): Set the commit message
        :param *args: Pass a non-keyworded, variable-length argument list to the function
        :param **kwargs: Pass additional parameters to the function
        :return: None
        """

        file_list = list(self.files.values())
        self.repo.upload_files(
            file_list, self.directory, commit_message=commit_message, *args, **kwargs
        )
        self._reset_dataset()



import yaml

from argparse import Namespace


def read_hparams(path: str) -> Namespace:
    """
    Use to read some hyperparameters which were previously saved as YAML.
    This is the format used by DAGsHubLogger, and the returned Namespace is compatible with pytorch-lightning models.
    """
    with open(path) as f:
        params = yaml.safe_load(f)
        return Namespace(**params)



import os.path
import pytest
from dagshub.streaming import DagsHubFilesystem


def test_sets_current_revision(mock_api):
    fs = DagsHubFilesystem()
    assert fs._current_revision == mock_api.current_revision
    assert mock_api["branch"].called


def test_open(mock_api, repo_with_hooks):
    path = "a.txt"
    content = b"Hello, streaming world!"
    read_route = mock_api.add_file(path, content)
    with open(path, "rb") as f:
        assert f.read() == content
    assert read_route.called


def test_open_nested_path(mock_api, repo_with_hooks):
    path = "nested/path/a.txt"
    content = b"Hello, streaming world!"
    mock_api.add_file(path, content)
    with open(path, "rb") as f:
        assert f.read() == content
    print(os.path.dirname(path))
    assert os.path.exists(os.path.dirname(path))


def test_open_for_write(mock_api, repo_with_hooks):
    content = "adfasdf"
    path = "new_file.txt"
    with open(path, "w") as f:
        f.write(content)
    with open(path, "r") as f:
        assert f.read() == content


def test_nested_open_for_write(mock_api, repo_with_hooks):
    content = "adfasdf"
    path = "aaaa/bbb/new_file.txt"
    mock_api.add_dir("aaaa", contents=[("bbb", "dir")])
    mock_api.add_dir("aaaa/bbb")
    with open(path, "w") as f:
        f.write(content)
    with open(path, "r") as f:
        assert f.read() == content


def test_nested_open_throws_nonexistent_dir(mock_api, repo_with_hooks):
    content = "adfasdf"
    path = "nonexistent/dir/new_file.txt"
    mock_api.add_dir("nonexistent", status=404)
    with pytest.raises(FileNotFoundError):
        with open(path, "w") as f:
            f.write(content)



import os

import pytest


def test_listdir(mock_api, repo_with_hooks):
    files = [("a.txt", "file"), ("b.txt", "file"), ("dir1", "dir")]
    path = "testdir"
    mock_api.add_dir(path, files)
    actual = os.listdir(path)
    expected = [f[0] for f in files]
    # Compare sets to ignore order of the lists
    assert set(actual) == set(expected)


def test_listdir_includes_local_files(mock_api, repo_with_hooks):
    path = "testdir"
    os.mkdir(path)
    with pytest.MonkeyPatch.context() as mp:
        mp.chdir(path)
        files = [("a.txt", "file"), ("b.txt", "file"), ("dir1", "dir")]
        mock_api.add_dir(path, files)
        add_file = "c.txt"
        with open(add_file, "w") as f:
            f.write("test")

        expected = [f[0] for f in files]
        expected.append(add_file)

        actual = os.listdir(".")
        assert set(expected) == set(actual)


def test_binary(mock_api, repo_with_hooks):
    files = [("a.txt", "file"), ("b.txt", "file"), ("dir1", "dir")]
    path = "testdir"
    mock_api.add_dir(path, files)
    actual = os.listdir(path.encode("utf-8"))
    expected = [f[0].encode("utf-8") for f in files]
    assert set(actual) == set(expected)



from fastai.learner import Learner, Recorder
from fastcore.basics import ignore_exceptions, patch, detuplify
from fastai.callback.core import Callback
from fastai.callback.hook import total_params
from fastai.torch_core import rank_distrib, to_detach

from ..logger import DAGsHubLogger as LoggerImpl


class DAGsHubLogger(Callback):
    """
    First, install the DAGsHub logger with
    `pip install dagshub`

    Example of Usage:
    ```
    from dagshub.fastai import DAGsHubLogger

    # To log only during a single training phase
    learn.fit(..., cbs=DAGsHubLogger())
    ```
    """

    "Saves model topology, losses & metrics"
    remove_on_fetch, order = True, Recorder.order + 1

    def __init__(self,
                 metrics_path: str = 'metrics.csv',
                 should_log_metrics: bool = True,
                 hparams_path: str = 'params.yml',
                 should_log_hparams: bool = True,
                 should_make_dirs: bool = True,
                 ):
        """
        :param metrics_path (str): Where to save the single metrics CSV file.
        :param should_log_metrics (bool): Whether to log metrics at all. Should probably always be True.
        :param hparams_path (str): Where to save the single hyperparameter YAML file.
        :param should_log_hparams (bool): Whether to log hyperparameters to a file.
            Should be False if you want to work with hyperparameters in a dependency file,
            rather than specifying them using command line arguments.
        :param should_make_dirs (bool): If true, the directory structure required by metrics_path
            and hparams_path will be created. Has no effect if the directory structure already exists.
        """
        super(DAGsHubLogger, self).__init__()
        self.logger = LoggerImpl(metrics_path=metrics_path, should_log_metrics=should_log_metrics,
                                 hparams_path=hparams_path, should_log_hparams=should_log_hparams,
                                 should_make_dirs=should_make_dirs, eager_logging=True)
        self._dags_status_hyperparam_name = 'success'
        self._dags_step_num = -1
        self._dags_epoch = 0

    def before_fit(self):
        # Make sure this is a training run
        self.run = not hasattr(self.learn, 'lr_finder') and \
                   not hasattr(self, "gather_preds") and rank_distrib() == 0
        if not self.run:
            return

        # Log Hyper Parameters
        params = self.learn.gather_args()
        self.logger.log_hyperparams(params)

    def after_batch(self):
        if self.learn.training:
            self._dags_step_num += 1
            self._dags_epoch += 1 / self.n_iter
            hypers = {f'{k}_{i}': v for i, h in enumerate(self.opt.hypers) for k, v in h.items()}
            metrics = {
                'train_loss': to_detach(self.learn.smooth_loss.clone()).numpy(),
                'raw_loss': to_detach(self.learn.loss.clone()).numpy(),
                'epoch': self._dags_epoch,
                **hypers
            }
            self.logger.log_metrics(metrics, step_num=self._dags_step_num)

    def after_epoch(self):
        self._dags_epoch = round(self._dags_epoch)
        metrics = {n: s for n, s in zip(self.recorder.metric_names, self.recorder.log) if n not
                   in ['train_loss', 'epoch', 'time']}
        metrics['epoch'] = self._dags_epoch
        self.logger.log_metrics(metrics, step_num=self._dags_step_num)

    def after_fit(self):
        self.run = True
        self._dags_epoch = round(self._dags_epoch)
        if self._dags_status_hyperparam_name is not None and \
                self._dags_status_hyperparam_name not in self.logger.hparams:
            self.logger.log_hyperparams({self._dags_status_hyperparam_name: self.run})
        metrics = {n: s for n, s in zip(self.recorder.metric_names, self.recorder.log) if n not
                   in ['train_loss', 'epoch', 'time']}
        metrics['epoch'] = self._dags_epoch
        self.logger.log_metrics(metrics, step_num=self._dags_step_num)


@patch
def gather_args(self: Learner):
    """Gather config parameters accessible to the learner"""
    # args stored by `store_attr`
    cb_args = {f'{cb}': getattr(cb, '__stored_args__', True) for cb in self.cbs}
    args = {'Learner': self, **cb_args}
    # input dimensions
    try:
        n_inp = self.dls.train.n_inp
        args['n_inp'] = n_inp
        xb = self.dls.train.one_batch()[:n_inp]
        args.update({f'input {n + 1} dim {i + 1}': d for n in range(n_inp) for i, d in
                     enumerate(list(detuplify(xb[n]).shape))})
    except Exception:
        print('Could not gather input dimensions')
    # other useful information
    with ignore_exceptions():
        args['batch size'] = self.dls.bs
        args['batch per epoch'] = len(self.dls.train)
        args['model parameters'] = total_params(self.model)[0]
        args['device'] = self.dls.device.type
        args['frozen'] = bool(self.opt.frozen_idx)
        args['frozen idx'] = self.opt.frozen_idx
        args['dataset.tfms'] = f'{self.dls.dataset.tfms}'
        args['dls.after_item'] = f'{self.dls.after_item}'
        args['dls.before_batch'] = f'{self.dls.before_batch}'
        args['dls.after_batch'] = f'{self.dls.after_batch}'
    return args



from dataclasses import dataclass

import httpx


@dataclass
class UploadErrorResponseContent:
    error: str
    details: str


@dataclass
class GenericAPIErrorContent:
    message: str


_error_lookup = {}


def register_upload_api_error(error_value: str):
    def decorator(cls):
        _error_lookup[error_value] = cls

        def __init__(self, details: str):
            self.details = details
        cls.__init__ = __init__
        return cls
    return decorator


@register_upload_api_error(error_value="missing last_commit")
class UpdateNotAllowedError(Exception):
    pass


@register_upload_api_error(error_value="invalid last_commit")
class InvalidLastCommitError(Exception):
    pass


@register_upload_api_error(error_value="versioning conflict")
class VersioningConflictError(Exception):
    pass


@register_upload_api_error(error_value="edit pipeline unsupported")
class UnsupportedStageContainerFileError(Exception):
    pass


@register_upload_api_error(error_value="path conflict")
class PathConflictError(Exception):
    pass


@register_upload_api_error(error_value="server error")
class InternalServerErrorError(Exception):
    pass


class DagsHubAPIError(Exception):
    """
    Generic API Exception, only has a message
    """
    def __init__(self, message: str):
        super().__init__()
        self.message = message


def determine_upload_api_error(response: httpx.Response) -> Exception:
    try:
        json_content = response.json()
    except Exception as e:
        return e

    if "error" in json_content and "details" in json_content:
        error_content = UploadErrorResponseContent(json_content["error"], json_content["details"])
        error_class = _error_lookup.get(error_content.error)
        if error_class is None:
            return DagsHubAPIError(f"{error_content.error}:\n{error_content.details}")
        return error_class(error_content.details)

    elif "message" in json_content:
        return DagsHubAPIError(json_content["message"])

    return RuntimeError(response.content)



import os
import pytest
from tests.dda.mock_api import MockApi


@pytest.fixture
def repouser():
    return "user"


@pytest.fixture
def reponame():
    return "repo"


@pytest.fixture
def repopath(repouser, reponame):
    return f"{repouser}/{reponame}"


@pytest.fixture
def current_revision(dagshub_repo):
    return dagshub_repo.api.heads.main.commit.hexsha


@pytest.fixture
def mock_api(dagshub_repo):
    with MockApi(
        git_repo=dagshub_repo, base_url="https://dagshub.com", assert_all_called=False
    ) as respx_mock:
        yield respx_mock


@pytest.fixture
def dagshub_repo(git_repo, repopath):
    cwd = os.getcwd()
    os.chdir(git_repo.workspace)
    repo = git_repo.api
    # Add remote
    repo.create_remote("origin", f"https://dagshub.com/{repopath}.git")
    # Create an initial commit
    filename = "test_git_file.txt"
    open(filename, "wb").close()
    repo.index.add([filename])
    repo.index.commit("initial commit")
    yield git_repo
    os.chdir(cwd)



import os
import pytest

import dagshub.streaming


def scandir_to_dict(scandir_iter):
    res = {}
    for f in scandir_iter:
        res[f.name] = f.path
    return res


@pytest.fixture
def scandir_mock(mock_api):
    mock_api.add_dir("temp", [("b.txt", "file")])
    yield mock_api


@pytest.mark.parametrize(
    "cwd, path",
    [
        (".", "temp"),
    ],
)
def test_scandir(scandir_mock, cwd, path):
    """
    Tests that DagsHubFilesystem behaves the same way as os.scandir()
    """
    # Assuming we only need to create one directory
    relpath = os.path.basename(path)
    os.mkdir(relpath)
    open(os.path.join(relpath, "a.txt"), "w").close()
    fs = dagshub.streaming.DagsHubFilesystem()
    try:
        with pytest.MonkeyPatch.context() as mp:
            mp.chdir(cwd)
            expected = scandir_to_dict(os.scandir(path))
            fs.install_hooks()
            actual = scandir_to_dict(os.scandir(path))
            assert "b.txt" in actual
            del actual["b.txt"]
            assert expected == actual
    finally:
        fs.uninstall_hooks()


def test_abspath(scandir_mock):
    # Need to do it separately instead of in parametrize because one of the fixtures changes the cwd
    test_scandir(scandir_mock, ".", os.path.abspath("temp"))


def test_nested(scandir_mock):
    repodir = os.path.basename(os.getcwd())
    test_scandir(scandir_mock, "..", f"{repodir}/temp")


def test_up(scandir_mock):
    repodir = os.path.basename(os.getcwd())
    test_scandir(scandir_mock, ".", f"../{repodir}/temp")



import configparser
import os
import urllib
from os.path import exists
from pathlib import Path

import git

from dagshub.auth import get_token
from dagshub.auth.token_auth import HTTPBearerAuth
from dagshub.common import config
from dagshub.common.helpers import get_project_root, http_request, log_message
from dagshub.upload import create_repo


def init(repo_name=None, repo_owner=None, url=None, root=None,
         host=config.host, mlflow=True, dvc=False):
    # Setup required variables
    if dvc:
        root = root or get_project_root(Path(os.path.abspath('.')))
        if not exists(root / '.git'):
            raise ValueError('No git project found! (stopped at mountpoint {root}). \
                              Please run this command in a git repository.')

    if url and (repo_name or repo_owner):
        repo_name, repo_owner = None, None

    if not url:
        if repo_name is not None and repo_owner is not None:
            url = urllib.parse.urljoin(host, f'{repo_owner}/{repo_name}')
        elif dvc:
            for remote in git.Repo(root).remotes:
                if host in remote.url:
                    url = remote.url[:-4]
    if not url:
        raise ValueError('No host remote found! Please specify the remote using the url variable, or --url argument.')
    elif url[-4] == '.':
        url = url[:-4]

    if not (repo_name and repo_owner):
        splitter = lambda x: (x[-1], x[-2]) # noqa E721
        repo_name, repo_owner = splitter(url.split('/'))

    if None in [repo_name, repo_owner, url]:
        raise ValueError('Could not parse repository owner and name. Make sure you specify either a link \
                          to the repository with --url or a pair of --repo-owner and --repo-name')

    # Setup authentication
    token = config.token or get_token()
    bearer = HTTPBearerAuth(token)

    # Configure repository
    res = http_request("GET", urllib.parse.urljoin(host, config.REPO_INFO_URL.format(
        owner=repo_owner,
        reponame=repo_name)), auth=bearer)
    if res.status_code == 404:
        create_repo(repo_name)

    # Configure MLFlow
    if mlflow:
        os.environ['MLFLOW_TRACKING_URI'] = f'{url}.mlflow'
        os.environ['MLFLOW_TRACKING_USERNAME'] = token
        os.environ['MLFLOW_TRACKING_PASSWORD'] = token

    # Configure DVC
    if dvc:
        Path(root / '.dvc').mkdir(parents=True, exist_ok=True)
        write = True

        dvc_config = configparser.ConfigParser()
        dvc_config_local = configparser.ConfigParser()
        dvc_config.read(root / '.dvc' / 'config')
        dvc_config_local.read(root / '.dvc' / 'config.local')

        for section in dvc_config.sections():
            if 'url' in dvc_config[section] and host in dvc_config[section]['url']:
                write = False

        remote = 'dagshub' if 'origin' in dvc_config.sections() else 'origin'
        if write:
            dvc_config_local[f'\'remote "{remote}"\''] = {'auth': 'basic',
                                                          'user': token,
                                                          'password': token}
            dvc_config[f'\'remote "{remote}"\''] = {'url': f'{url}.dvc'}

            with open(root / '.dvc' / 'config', 'w') as config_file, \
                 open(root / '.dvc' / 'config.local', 'w') as config_local_file:
                dvc_config.write(config_file)
                dvc_config_local.write(config_local_file)
                log_message(f'Added new remote "{remote}" with url = {url}')

        if not exists(root / '.dvc' / '.gitignore'):
            with open(root / '.dvc' / '.gitignore', 'w') as config_gitignore:
                config_gitignore.write(config.CONFIG_GITIGNORE)

    log_message('Repository initialized!')



import typing
from httpx import Request, Response, Auth


class HTTPBearerAuth(Auth):
    """Attaches HTTP Bearer Authorization to the given Request object."""

    def __init__(self, token):
        self.token = token

    def auth_flow(self, request: Request) -> typing.Generator[Request, Response, None]:
        request.headers["Authorization"] = f"Bearer {self.token}"
        yield request

    def __eq__(self, other):
        return all([
            self.token == getattr(other, 'token', None),
            ])

    def __ne__(self, other):
        return not self == other

    def __call__(self, r):
        r.headers['Authorization'] = f'Bearer {self.token}'
        return r



import builtins
import io
import os
import re
import subprocess
import sys
from configparser import ConfigParser
from contextlib import contextmanager
from functools import partial, wraps, lru_cache
from multiprocessing import AuthenticationError
from os import PathLike
from pathlib import Path
from typing import Optional, TypeVar, Union, Dict, Set
from urllib.parse import urlparse
from dagshub.common import config, helpers
import logging
from dagshub.common.helpers import http_request, get_project_root

# Pre 3.11 - need to patch _NormalAccessor for _pathlib, because it pre-caches open and other functions.
# In 3.11 _NormalAccessor was removed
PRE_PYTHON3_11 = sys.version_info.major == 3 and sys.version_info.minor < 11
if PRE_PYTHON3_11:
    from pathlib import _NormalAccessor as _pathlib  # noqa: E402

T = TypeVar('T')
logger = logging.getLogger(__name__)


def wrapreturn(wrappertype):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return wrappertype(func(*args, **kwargs))

        return wrapper

    return decorator


class dagshub_ScandirIterator:
    def __init__(self, iterator):
        self._iterator = iterator

    def __iter__(self):
        return self._iterator

    def __next__(self):
        return self._iterator.__next__()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        return self


def cache_by_path(func):
    cache = {}

    @wraps(func)
    def wrapper(self, path: str, include_size: bool = False):
        if not include_size and (path, True) in cache:
            cache[path, False] = cache[path, True]
        if (path, include_size) not in cache:
            cache[path, include_size] = func(self, path, include_size)
        return cache[path, include_size]

    wrapper.cache = cache
    return wrapper


SPECIAL_FILE = Path('.dagshub-streaming')


# TODO: Singleton metaclass that lets us keep a "main" DvcFilesystem instance
class DagsHubFilesystem:
    """
    A DagsHub-repo aware filesystem class

    :param project_root: Path to the git repository with the repo.
        If None, we look up the filesystem from the current dir until we find a git repo
    :param repo_url: URL to the DagsHub repository.
        If None, URL is received from the git configuration
    :param branch: Explicitly sets a branch/commit revision to work with
        If None, branch is received from the git configuration
    :param username: DagsHub username
    :param password: DagsHub password
    :param token: DagsHub API token (as an alternative login variant to username/password)
    :param timeout: Timeout in seconds for HTTP requests.
        Influences all requests except for file download, which has no timeout
    """

    __slots__ = ('project_root',
                 'project_root_fd',
                 'dvc_remote_url',
                 'user_specified_branch',
                 'parsed_repo_url',
                 'remote_tree',
                 'username',
                 'password',
                 'dagshub_remotes',
                 'token',
                 'timeout',
                 '__weakref__')

    def __init__(self,
                 project_root: Optional[PathLike] = None,
                 repo_url: Optional[str] = None,
                 branch: Optional[str] = None,
                 username: Optional[str] = None,
                 password: Optional[str] = None,
                 token: Optional[str] = None,
                 timeout: Optional[int] = None,
                 _project_root_fd: Optional[int] = None):

        # Find root directory of Git project
        if not project_root:
            self.project_root = get_project_root(Path(os.path.abspath('.')))
        else:
            self.project_root = Path(os.path.abspath(project_root))
        del project_root
        # TODO: if no Git project found, search for .dvc project?

        if _project_root_fd:
            self.project_root_fd = _project_root_fd
        else:
            self.project_root_fd = os.open(self.project_root, os.O_DIRECTORY)

        self.dagshub_remotes = []
        self.parse_git_config()

        if not repo_url:
            if len(self.dagshub_remotes) > 0:
                repo_url = self.dagshub_remotes[0]
            else:
                raise ValueError('No DagsHub git remote detected, please specify repo_url= argument or --repo_url flag')

        self.user_specified_branch = branch
        self.parsed_repo_url = urlparse(repo_url)
        self.dvc_remote_url = f'{repo_url}.dvc/cache'
        # Key: path, value: dict of {name, type} on that path (in remote)
        self.remote_tree: Dict[str, Dict[str, str]] = {}

        # Determine if any authentication is needed
        self.username = username or config.username
        self.password = password or config.password
        self.token = token or config.token
        self.timeout = timeout or config.http_timeout

        response = self._api_listdir('')
        if response.status_code < 400:
            pass
        else:
            # TODO: Check .dvc/config{,.local} for credentials
            raise AuthenticationError('DagsHub credentials required, however none provided or discovered')

    @property
    @lru_cache(maxsize=None)
    def _current_revision(self) -> str:
        """
        Gets current revision on repo:
        - If User specified a branch, returns HEAD of that brunch on the remote
        - If branch wasn't detected, returns HEAD of default branch in the speficied remote.
        - If HEAD is a branch, tries to find a dagshub remote associated with it and get its HEAD
        - If HEAD is a commit revision, checks that the commit exists on DagsHub
        """

        if self.user_specified_branch:
            branch = self.user_specified_branch
        else:
            try:
                with open(self.project_root / ".git/HEAD") as head_file:
                    head = head_file.readline().strip()
                if head.startswith("ref"):
                    branch = head.split("/")[-1]
                else:
                    # contents of HEAD is the revision - check that this commit exists on remote
                    if self.is_commit_on_remote(head):
                        return head
                    else:
                        raise RuntimeError(f"Current HEAD ({head}) doesn't exist on the remote. "
                                           f"Please push your changes to the remote or checkout a tracked branch.")

            except FileNotFoundError:
                logger.debug("Couldn't get branch info from local git repository, " +
                             "fetching default branch from the remote...")
                owner, reponame = self.parsed_repo_url.path.split("/")[1:]
                branch = helpers.get_default_branch(owner, reponame, self.auth)
                logger.debug(f'Set default branch: "{branch}"')
        return self.get_remote_branch_head(branch)

    @property
    def content_api_url(self):
        return self.get_api_url(f"/api/v1/repos{self.parsed_repo_url.path}/content/{self._current_revision}")

    @property
    def raw_api_url(self):
        return self.get_api_url(f"/api/v1/repos{self.parsed_repo_url.path}/raw/{self._current_revision}")

    def is_commit_on_remote(self, sha1):
        url = self.get_api_url(f"/api/v1/repos{self.parsed_repo_url.path}/commits/{sha1}")
        resp = self.http_get(url)
        return resp.status_code == 200

    def get_remote_branch_head(self, branch):
        url = self.get_api_url(f"/api/v1/repos{self.parsed_repo_url.path}/branches/{branch}")
        resp = self.http_get(url)
        if resp.status_code != 200:
            raise RuntimeError(f"Got status {resp.status_code} while trying to get head of branch {branch}. \r\n"
                               f"Response body: {resp.content}")
        return resp.json()["commit"]["id"]

    def get_api_url(self, path):
        return str(self.parsed_repo_url._replace(path=path).geturl())

    @property
    def auth(self):
        import dagshub.auth
        from dagshub.auth.token_auth import HTTPBearerAuth

        if self.username is not None and self.password is not None:
            return self.username, self.password

        try:
            token = self.token or dagshub.auth.get_token()
        except dagshub.auth.OauthNonInteractiveShellException:
            logger.debug("Failed to perform OAuth in a non interactive shell")
        if token is not None:
            return HTTPBearerAuth(token)

        # Try to fetch credentials from the git credential file
        proc = subprocess.run(['git', 'credential', 'fill'],
                              input=f'url={self.repo_url}'.encode(),
                              capture_output=True)
        answer = {line[:line.index('=')]: line[line.index('=') + 1:]
                  for line in proc.stdout.decode().splitlines()}
        if 'username' in answer and 'password' in answer:
            return answer['username'], answer['password']

    def parse_git_config(self):
        # Get URLs of dagshub remotes
        git_config = ConfigParser()
        git_config.read(Path(self.project_root) / '.git/config')
        git_remotes = [urlparse(git_config[remote]['url'])
                       for remote in git_config
                       if remote.startswith('remote ')]
        for remote in git_remotes:
            if remote.hostname != config.hostname:
                continue
            remote = remote._replace(netloc=remote.hostname)
            remote = remote._replace(path=re.compile(r'(\.git)?/?$').sub('', remote.path))
            self.dagshub_remotes.append(remote.geturl())

    def __del__(self):
        os.close(self.project_root_fd)

    def _relative_path(self, file: Union[str, PathLike, int]):
        if isinstance(file, int):
            return None
        if file == "":
            return None
        path = os.path.abspath(file)
        try:
            rel = Path(path).relative_to(os.path.abspath(self.project_root))
            if str(rel).startswith("<"):
                return None
            return rel
        except ValueError:
            return None

    @staticmethod
    def _passthrough_path(relative_path: PathLike):
        str_path = str(relative_path)
        if "/site-packages/" in str_path:
            return True
        return str_path.startswith(('.git/', '.dvc/')) or str_path in (".git", ".dvc")

    def _special_file(self):
        # TODO Include more information in this file
        return b'v0\n'

    def open(self, file, mode='r', buffering=-1, encoding=None,
             errors=None, newline=None, closefd=True, opener=None):
        if type(file) is bytes:
            file = os.fsdecode(file)
        relative_path = self._relative_path(file)
        if relative_path:
            if opener is not None:
                raise NotImplementedError('DagsHub\'s patched open() does not support custom openers')
            project_root_opener = partial(os.open, dir_fd=self.project_root_fd)
            if self._passthrough_path(relative_path):
                return self.__open(relative_path, mode, buffering, encoding, errors, newline,
                                   closefd, opener=project_root_opener)
            elif relative_path == SPECIAL_FILE:
                return io.BytesIO(self._special_file())
            else:
                try:
                    return self.__open(relative_path, mode, buffering, encoding, errors, newline,
                                       closefd, opener=project_root_opener)
                except FileNotFoundError as err:
                    # Open for reading - try to download the file
                    if "r" in mode:
                        resp = self._api_download_file_git(relative_path)
                        if resp.status_code < 400:
                            self._mkdirs(relative_path.parent, dir_fd=self.project_root_fd)
                            # TODO: Handle symlinks
                            with self.__open(relative_path, 'wb', opener=project_root_opener) as output:
                                output.write(resp.content)
                            return self.__open(relative_path, mode, buffering, encoding, errors, newline,
                                               closefd, opener=project_root_opener)
                        else:
                            # TODO: After API no longer 500s on FileNotFounds
                            #       check status code and only return FileNotFound on 404s
                            raise FileNotFoundError(f'Error finding {relative_path} in repo or on DagsHub')
                    # Write modes - make sure that the folder is a tracked folder (create if doesn't exist on disk),
                    # and then let the user write to file
                    else:
                        try:
                            # Using the fact that stat creates tracked dirs (but still throws on nonexistent dirs)
                            _ = self.stat(self.project_root / relative_path.parent)
                        except FileNotFoundError:
                            raise err
                        # Try to download the file if we're in append modes
                        if "a" in mode or "+" in mode:
                            resp = self._api_download_file_git(relative_path)
                            if resp.status_code < 400:
                                with self.__open(relative_path, 'wb', opener=project_root_opener) as output:
                                    output.write(resp.content)
                        return self.__open(relative_path, mode, buffering, encoding, errors, newline,
                                           closefd, opener=project_root_opener)

        else:
            return self.__open(file, mode, buffering, encoding, errors, newline,
                               closefd, opener)

    def os_open(self, path, flags, mode=0o777, *, dir_fd=None):
        """
        os.open is supposed to be lower level, but it's still being used by e.g. Pathlib
        We're trying to wrap around it here, by parsing flags and calling the higher-level open
        Caveats: list of flags being handled is not exhaustive + mode doesn't work
                 (because we lose them when passing to builtin open)
        WARNING: DO NOT patch actual os.open with it, because the builtin uses os.open.
                 This is only for the purposes of patching pathlib.open in Python 3.9 and below.
                 Since Python 3.10 pathlib uses io.open, and in Python 3.11 they removed the accessor completely
        """
        if dir_fd is not None:   # If dir_fd supplied, path is relative to that dir's fd, will handle in the future
            logger.debug("fs.os_open - NotImplemented")
            raise NotImplementedError('DagsHub\'s patched os.open() (for pathlib only) does not support dir_fd')
        if self._relative_path(path):
            try:
                open_mode = "r"
                # Write modes - calling in append mode,
                # This way we create the intermediate folders if file doesn't exist, but the folder it's in does
                # Append so we don't truncate the file
                if not (flags & os.O_RDONLY):
                    open_mode = "a"
                logger.debug("fs.os_open - trying to materialize path")
                self.open(path, mode=open_mode).close()
                logger.debug("fs.os_open - successfully materialized path")
            except FileNotFoundError:
                logger.debug("fs.os_open - failed to materialize path, os.open will throw")
        return os.open(path, flags, mode, dir_fd=dir_fd)

    def stat(self, path, *, dir_fd=None, follow_symlinks=True):
        if type(path) is bytes:
            path = os.fsdecode(path)
        if dir_fd is not None or not follow_symlinks:
            logger.debug("fs.stat - NotImplemented")
            raise NotImplementedError('DagsHub\'s patched stat() does not support dir_fd or follow_symlinks')
        relative_path = self._relative_path(path)
        # todo: remove False
        if relative_path:
            logger.debug("fs.stat - is relative path")
            if self._passthrough_path(relative_path):
                return self.__stat(relative_path, dir_fd=self.project_root_fd)
            elif relative_path == SPECIAL_FILE:
                return dagshub_stat_result(self, path, is_directory=False, custom_size=len(self._special_file()))
            else:
                try:
                    logger.debug(f"fs.stat - calling __stat - relative_path: {path}, dir_fd: {self.project_root_fd}")
                    return self.__stat(relative_path, dir_fd=self.project_root_fd)
                except FileNotFoundError as err:
                    logger.debug("fs.stat - FileNotFoundError")
                    logger.debug(f"remote_tree: {self.remote_tree}")
                    parent_path = relative_path.parent
                    if str(parent_path) not in self.remote_tree:
                        try:
                            # Run listdir to update cache
                            self.listdir(self.project_root / parent_path)
                        except FileNotFoundError:
                            raise err

                    cached_remote_parent_tree = self.remote_tree.get(str(parent_path))
                    logger.debug(f"cached_remote_parent_tree: {cached_remote_parent_tree}")

                    if cached_remote_parent_tree is None:
                        raise err

                    filetype = cached_remote_parent_tree.get(relative_path.name)
                    if filetype is None:
                        raise err

                    if filetype == "file":
                        return dagshub_stat_result(self, path, is_directory=False)
                    elif filetype == "dir":
                        self._mkdirs(relative_path, dir_fd=self.project_root_fd)
                        return self.__stat(relative_path, dir_fd=self.project_root_fd)
                    else:
                        raise RuntimeError(f"Unknown file type {filetype} for path {str(relative_path)}")

        else:
            return self.__stat(path, follow_symlinks=follow_symlinks)

    def chdir(self, path):
        if type(path) is bytes:
            path = os.fsdecode(path)
        relative_path = self._relative_path(path)
        if relative_path:
            abspath = os.path.join(self.project_root, relative_path)
            try:
                self.__chdir(abspath)
            except FileNotFoundError:
                resp = self._api_listdir(relative_path)
                # FIXME: if path is file, return FileNotFound instead of the listdir error
                if resp.status_code < 400:
                    self._mkdirs(relative_path, dir_fd=self.project_root_fd)
                    self.__chdir(abspath)
                else:
                    raise
        else:
            self.__chdir(path)

    def listdir(self, path='.'):
        # listdir needs to return results for bytes path arg also in bytes
        is_bytes_path_arg = type(path) is bytes
        if is_bytes_path_arg:
            str_path = os.fsdecode(path)
        else:
            str_path = path
        relative_path = self._relative_path(str_path)
        if relative_path:
            if self._passthrough_path(relative_path):
                with self._open_fd(relative_path) as fd:
                    return self.__listdir(fd)
            else:
                dircontents: Set[str] = set()
                error = None
                try:
                    with self._open_fd(relative_path) as fd:
                        dircontents.update(self.__listdir(fd))
                except FileNotFoundError as e:
                    error = e
                if relative_path == Path():
                    dircontents.add(SPECIAL_FILE.name)
                resp = self._api_listdir(relative_path)
                if resp.status_code < 400:
                    dircontents.update(Path(f['path']).name for f in resp.json())
                    self.remote_tree[str(relative_path)] = {
                        Path(f["path"]).name: f["type"]
                        for f in resp.json()
                    }
                    res = list(dircontents)
                    if is_bytes_path_arg:
                        res = [os.fsencode(p) for p in res]
                    return res
                else:
                    if error is not None:
                        raise error
                    else:
                        res = list(dircontents)
                        if is_bytes_path_arg:
                            res = [os.fsencode(p) for p in res]
                        return res
        else:
            return self.__listdir(path)

    @wrapreturn(dagshub_ScandirIterator)
    def scandir(self, path='.'):
        # scandir needs to return name and path as bytes, if entry arg is bytes
        is_bytes_path_arg = type(path) is bytes
        if is_bytes_path_arg:
            str_path = os.fsdecode(path)
        else:
            str_path = path
        relative_path = self._relative_path(str_path)
        if relative_path and not self._passthrough_path(relative_path):
            path = Path(str_path)
            local_filenames = set()
            try:
                for direntry in self.__scandir(path):
                    local_filenames.add(direntry.name)
                    yield direntry
                if relative_path == Path():
                    if SPECIAL_FILE.name not in local_filenames:
                        yield dagshub_DirEntry(self, path / SPECIAL_FILE,
                                               is_directory=False, is_binary=is_bytes_path_arg)
            except FileNotFoundError:
                pass
            resp = self._api_listdir(relative_path)
            if resp.status_code < 400:
                for f in resp.json():
                    name = Path(f['path']).name
                    if name not in local_filenames:
                        yield dagshub_DirEntry(self, path / name, f['type'] == 'dir', is_binary=is_bytes_path_arg)
        else:
            for entry in self.__scandir(path):
                yield entry

    @cache_by_path
    def _api_listdir(self, path: str, include_size: bool = False):
        return self.http_get(f'{self.content_api_url}/{path}',
                             params={'include_size': 'true'} if include_size else {},
                             headers=config.requests_headers)

    def _api_download_file_git(self, path: str):
        return self.http_get(f'{self.raw_api_url}/{path}', headers=config.requests_headers, timeout=None)

    def http_get(self, path: str, **kwargs):
        timeout = self.timeout
        if "timeout" in kwargs:
            timeout = kwargs["timeout"]
            del kwargs["timeout"]
        return http_request("GET", path, auth=self.auth, timeout=timeout, **kwargs)

    def install_hooks(self):
        if not hasattr(self.__class__, f'_{self.__class__.__name__}__unpatched'):
            # TODO: DRY this dictionary. i.e. __open() links cls.__open
            #  and io.open even though this dictionary links them
            #  Cannot use a dict as the source of truth because type hints rely on
            #  __get_unpatched inferring the right type
            self.__class__.__unpatched = {
                'open': io.open,
                'stat': os.stat,
                'listdir': os.listdir,
                'scandir': os.scandir,
                'chdir': os.chdir,
            }
            if PRE_PYTHON3_11:
                self.__class__.__unpatched["pathlib_open"] = _pathlib.open
        io.open = builtins.open = self.open
        os.stat = self.stat
        os.listdir = self.listdir
        os.scandir = self.scandir
        os.chdir = self.chdir
        if PRE_PYTHON3_11:
            if sys.version_info.minor == 10:
                # Python 3.10 - pathlib uses io.open
                _pathlib.open = self.open
            else:
                # Python <=3.9 - pathlib uses os.open
                _pathlib.open = self.os_open
            _pathlib.stat = self.stat
            _pathlib.listdir = self.listdir
            _pathlib.scandir = self.scandir

        self.__class__.hooked_instance = self

    @classmethod
    def uninstall_hooks(cls):
        if hasattr(cls, f'_{cls.__name__}__unpatched'):
            io.open = builtins.open = cls.__unpatched['open']
            os.stat = cls.__unpatched['stat']
            os.listdir = cls.__unpatched['listdir']
            os.scandir = cls.__unpatched['scandir']
            os.chdir = cls.__unpatched['chdir']
            if PRE_PYTHON3_11:
                _pathlib.open = cls.__unpatched['pathlib_open']
                _pathlib.stat = cls.__unpatched['stat']
                _pathlib.listdir = cls.__unpatched['listdir']
                _pathlib.scandir = cls.__unpatched['scandir']

    def _mkdirs(self, relative_path: PathLike, dir_fd: Optional[int] = None):
        for parent in list(relative_path.parents)[::-1]:
            try:
                self.__stat(parent, dir_fd=dir_fd)
            except (OSError, ValueError):
                os.mkdir(parent, dir_fd=dir_fd)
        try:
            self.__stat(relative_path, dir_fd=dir_fd)
        except (OSError, ValueError):
            os.mkdir(relative_path, dir_fd=dir_fd)

    @contextmanager
    def _open_fd(self, relative_path):
        fd = None
        try:
            fd = os.open(relative_path, os.O_DIRECTORY, dir_fd=self.project_root_fd)
            yield fd
        finally:
            if fd is not None:
                os.close(fd)

    @classmethod
    def __get_unpatched(cls, key, alt: T) -> T:
        if hasattr(cls, f'_{cls.__name__}__unpatched'):
            return cls.__unpatched[key]
        else:
            return alt

    @property
    def __open(self):
        return self.__get_unpatched('open', io.open)

    @property
    def __stat(self):
        return self.__get_unpatched('stat', os.stat)

    @property
    def __listdir(self):
        return self.__get_unpatched('listdir', os.listdir)

    @property
    def __scandir(self):
        return self.__get_unpatched('scandir', os.scandir)

    @property
    def __chdir(self):
        return self.__get_unpatched("chdir", os.chdir)


def install_hooks(project_root: Optional[PathLike] = None,
                  repo_url: Optional[str] = None,
                  branch: Optional[str] = None,
                  username: Optional[str] = None,
                  password: Optional[str] = None,
                  token: Optional[str] = None,
                  timeout: Optional[int] = None):
    """
    Monkey patches builtin Python functions to make them DagsHub-repo aware.
    Patched functions are: `open()`, `os.listdir()`, `os.scandir()`, `os.stat()` + pathlib's functions that use them

    This is equivalent to creating a `DagsHubFilesystem` object and calling its `install_hooks()` method

    :param project_root: Path to the git repository with the repo.
        If None, we look up the filesystem from the current dir until we find a git repo
    :param repo_url: URL to the DagsHub repository.
        If None, URL is received from the git configuration
    :param branch: Explicitly sets a branch/commit revision to work with
        If None, branch is received from the git configuration
    :param username: DagsHub username
    :param password: DagsHub password
    :param token: DagsHub API token (as an alternative login variant to username/password)
    :param timeout: Timeout in seconds for HTTP requests.
        Influences all requests except for file download, which has no timeout
    """
    fs = DagsHubFilesystem(project_root=project_root, repo_url=repo_url, branch=branch, username=username,
                           password=password, token=token, timeout=timeout)
    fs.install_hooks()


def uninstall_hooks():
    DagsHubFilesystem.uninstall_hooks()


class dagshub_stat_result:
    def __init__(self, fs: 'DagsHubFilesystem', path: PathLike, is_directory: bool, custom_size: int = None):
        self._fs = fs
        self._path = path
        self._is_directory = is_directory
        self._custom_size = custom_size
        assert not self._is_directory  # TODO make folder stats lazy?

    def __getattr__(self, name: str):
        if not name.startswith('st_'):
            raise AttributeError
        if hasattr(self, '_true_stat'):
            return os.stat_result.__getattribute__(self._true_stat, name)
        if name == 'st_uid':
            return os.getuid()
        elif name == 'st_gid':
            return os.getgid()
        elif name == 'st_atime' or name == 'st_mtime' or name == 'st_ctime':
            return 0
        elif name == 'st_mode':
            return 0o100644
        elif name == 'st_size':
            if self._custom_size:
                return self._custom_size
            return 1100  # hardcoded size because size requests take a disproportionate amount of time
        self._fs.open(self._path)
        self._true_stat = self._fs._DagsHubFilesystem__stat(self._fs._relative_path(self._path),
                                                            dir_fd=self._fs.project_root_fd)
        return os.stat_result.__getattribute__(self._true_stat, name)

    def __repr__(self):
        inner = repr(self._true_stat) if hasattr(self, '_true_stat') else 'pending...'
        return f'dagshub_stat_result({inner}, path={self._path})'


class dagshub_DirEntry:
    def __init__(self, fs: 'DagsHubFilesystem', path: PathLike, is_directory: bool = False, is_binary: bool = False):
        self._fs = fs
        self._path = path
        self._is_directory = is_directory
        self._is_binary = is_binary

    @property
    def name(self):
        # TODO: create decorator for delegation
        if hasattr(self, '_true_direntry'):
            name = self._true_direntry.name
        else:
            name = self._path.name
        return os.fsencode(name) if self._is_binary else name

    @property
    def path(self):
        if hasattr(self, '_true_direntry'):
            path = self._true_direntry.path
        else:
            path = str(self._path)
        return os.fsencode(path) if self._is_binary else path

    def is_dir(self):
        if hasattr(self, '_true_direntry'):
            return self._true_direntry.is_dir()
        else:
            return self._is_directory

    def is_file(self):
        if hasattr(self, '_true_direntry'):
            return self._true_direntry.is_file()
        else:
            # TODO: Symlinks should return false
            return not self._is_directory

    def stat(self):
        if hasattr(self, '_true_direntry'):
            return self._true_direntry.stat()
        else:
            return self._fs.stat(self._path)

    def __getattr__(self, name: str):
        if name == '_true_direntry':
            raise AttributeError
        if hasattr(self, '_true_direntry'):
            return os.DirEntry.__getattribute__(self._true_direntry, name)
        if self._is_directory:
            self._fs._mkdirs(self._fs._relative_path(self._path), dir_fd=self._fs.project_root_fd)
        else:
            self._fs.open(self._path)
        with self._open_fd(self._fs._relative_path(self._path).parent) as fd:
            for direntry in self._fs._DagsHubFilesystem__scandir(fd):
                if direntry.name == self._path.name:
                    self._true_direntry = direntry
                    return os.DirEntry.__getattribute__(self._true_direntry, name)
            else:
                raise FileNotFoundError

    def __repr__(self):
        cached = ' (cached)' if hasattr(self, '_true_direntry') else ''
        return f'<dagshub_DirEntry \'{self.name}\'{cached}>'


# Used for testing purposes only
if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    fs = DagsHubFilesystem()
    fs.install_hooks()

__all__ = [DagsHubFilesystem.__name__, install_hooks.__name__]
